/*
 * Log.cpp
 *
 *  Created on: 26 feb 2015
 *      Author: Daniel
 */

#include "Log.h"

namespace infomap {

unsigned int Log::s_verboseLevel = 0;
bool Log::s_silent = false;

} // namespace infomap
